#!/usr/bin/env python3

# SECLAF Sequence Classification Framework
# Copyright (c) 2016-2017 Balazs Szalkai

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os,re,math,time,sys,random
import gzip

# annotation.py: Annotation file reading and writing utilities.
# An annotation file is a collection of sequences and corresponding classes.
# These files can be used for training and testing networks.

# Class for writing sequence annotations (classifications).
class AnnotationWriter:
	# Constructs an AnnotationWriter which will write to the specified file in .ann or .ann.gz format.
	# filename: an annotation file (text file), or a gzipped annotation file (name must end in .gz then)
	def __init__(self, filename):
		self.filename = filename
		self.f = gzip.open(filename, "wt") if filename.endswith(".gz") else open(filename, "wt")
		self.f.write("// Annotation file (sequence, classes)\n")
		
	# Closes the file associated with this AnnotationWriter.
	# If already closed, does nothing.
	def close(self):
		if self.f is not None:
			self.f.close()
			self.f = None
	
	# Writes a sequence.
	# seq: a sequence
	# classification_list: the list of classes seq belongs to (may also be an empty list)
	# description: any (human-readable) description text for seq
	def write(self, seq, classification_list, description = ""):
		assert self.f is not None, "Annotation writer already closed"
		assert seq.find("\t") < 0
		assert isinstance(classification_list, list)
		line = seq+"\t"+("&".join(classification_list))
		if description != "":
			line += "\t"+description
		self.f.write(line+"\n")

	# Destructor. Automatically closes the writer if not already closed.
	def __del__(self):
		self.close()
		
# Class for reading a collection of sequences along with their annotations (classifications).
# Input can be a FASTA file or an .ann file (own format).
# Input can also be gzipped (in this case, filename must end in .gz).
class AnnotationReader:
	# Constructs an AnnotationReader which will read from the specified file.
	# filename: the name of an .ann, .ann.gz, .fasta or .fasta.gz file.
	# minLength: minimum valid sequence length (shorter sequences are discarded)
	# maxLength: maximum valid sequence length (longer sequences are discarded)
	# removeStartingM: whether the leading M amino acid should be removed from the sequences. Must be True or False.
	#                  In some files, all sequences begin with M, which is a false Methionine and in fact corresponds to the start codon.
	#                  In this case, removeStartingM must be True.
	def __init__(self, filename, minLength, maxLength, removeStartingM):
		self.filename = filename
		self.f = gzip.open(filename, "rt") if filename.endswith(".gz") else open(filename, "rt")
		self.minLength = minLength
		self.maxLength = maxLength
		self.removeStartingM = removeStartingM
		
		# FASTA file auto-detection and read state
		self.isFasta = None
		self.fasta_seq = ""
		self.fasta_classes = ""
	
	# Closes the file associated with this AnnotationReader.
	# If already closed, does nothing.
	def close(self):
		if self.f is not None:
			self.f.close()
			self.f = None
			
	# Destructor. Automatically closes the reader if not already closed.
	def __del__(self):
		self.close()

	# Moves the reader before the first sequence, as if the file was freshly opened.
	def rewind(self):
		self.f.seek(0)

	# Internal postprocessing after reading a sequence.
	# Returns a tuple (seq, classes) on success; but returns None if sequence is too short or long.
	def _postprocess(self, seq, classes):
		if self.removeStartingM and seq.startswith("M"):
			seq = seq[1:]
		if len(seq) >= self.minLength and len(seq) <= self.maxLength:
			classes = [] if classes in ["", "-"] else classes.split("&")
			return seq, classes
		else:
			return None

	# Reads a sequence from the annotation file.
	# Returns: a tuple of (sequence (string), annotations/classes (list of strings)).
	# By default, returns (None, None) if file ended. However, if autorewind is True, proceeds with rewinding and returns the first valid sequence in the file.
	def read(self, autorewind=False):
		while True:
			line = self.f.readline()
			eof = not line
			# do not signal eof if last sequence is still to be read
			if eof and (not self.isFasta or self.fasta_seq == ""):
				if autorewind:
					self.rewind()
					continue
				else:
					return None, None

			if line is None:
				line = ""
			elif line.startswith("//") or line.startswith("#"):
				continue
			
			line = line.rstrip()
			if self.isFasta is None:
				self.isFasta = line.startswith(">")
				
			if self.isFasta:
				if line.startswith(">") or eof:
					# if we got here with eof, we should have a valid self.fasta_seq
					result = self._postprocess(self.fasta_seq, self.fasta_classes)
					
					# clear FASTA state
					self.fasta_seq = ""
					classes_str = "classes="
					classes_start = line.lower().find(classes_str)
					self.fasta_classes = "" if classes_start < 0 else line[classes_start + len(classes_str):]
					
					if result is not None:
						return result
				else:
					self.fasta_seq += line
			else: # .ann format
				tok = line.split("\t")
				seq = tok[0]
				result = self._postprocess(seq, tok[1])
				if result is not None:
					return result
	
# Wraps an AnnotationReader and returns annotations a bit shuffled.
# Will own the provided AnnotationReader (it will also be closed if the shuffler is closed).
# Shuffling can greatly improve training efficiency, but the annotation input file must already be shuffled,
# because AnnotationShuffler has a limited efficiency (does not shuffle ALL the sequences in the file at once, just a range of them).
class AnnotationShuffler:
	# Constructs the shuffler around the given reader.
	# Afterwards the supplied reader need not be closed, closing the shuffler does it automatically.
	# annotationReader: an AnnotationReader (or compatible) instance
	# bufferSize: size of shuffling buffer. This many sequences are read at once, then shuffled.
	#             If too large, a lot of memory will be consumed. If small, shuffling will not be efficient.
	def __init__(self, annotationReader, bufferSize = 32768):
		self.ar = annotationReader
		self.buffer = []
		self.bufferSize = bufferSize
	
	# Closes the underlying reader.
	def close(self):
		if self.ar is not None:
			self.ar.close()
			self.ar = None
	
	# Destructor. Does close() automatically.
	def __del__(self):
		self.close()
		
	# Reads a sequence from the annotation input file.
	# Returns a tuple of (sequence (string), annotations/classes (list of strings)).
	# Always returns a valid data point (automatically rewinds the underlying reader if necessary).
	def read(self):
		if len(self.buffer) < self.bufferSize:
			# do not fill the whole buffer at once, this way we dont have to wait a lot
			# only 8 items at a time
			for i in range(min(8, self.bufferSize - len(self.buffer))):
				data = self.ar.read()
				if data[0] is None: # end of reader?
					if len(self.buffer) == 0:
						self.ar.rewind()
						continue
					else:
						break # first let the buffer go empty (avoid surviving items through multiple rewinds, they may duplicate)
				else:
					self.buffer.append(data)
		index = random.randint(0, len(self.buffer)-1)
		result = self.buffer[index]
		self.buffer[index] = self.buffer[-1]
		self.buffer.pop()
		return result
		