#!/usr/bin/env python3

# SECLAF Sequence Classification Framework
# Copyright (c) 2016-2017 Balazs Szalkai

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os,re,math,time,sys
import json
from . import encode_x

# config.py: When included, parses the config.json file, which contains most of the arguments and parameters for SECLAF.
# Please see readme.html for description of parameters in config.json.
# Some configuration parameters are required, others are optional (where a default value is given).
# The parsed configuration params are accessible as config.param_name (i.e. as globals of the "config" module).

fnConfigJson = sys.argv[1]
dirConfigJson = os.path.dirname(os.path.abspath(fnConfigJson))
if not dirConfigJson.endswith("/"):
	dirConfigJson += "/"
with open(fnConfigJson) as jsonData:
	configJson = json.load(jsonData)
	print(fnConfigJson+" successfully loaded")

configDesc = {}
def getConfig(id, defaultValue = None, help = None):
	assert help is not None
	configDesc[id] = {"defaultValue":defaultValue, "help":help}
	if id in configJson:
		return configJson[id]
	else:
		if defaultValue is None:
			print("'%s' is a required parameter, must be set to a value in %s" % (id, fnConfigJson))
			print("%s: %s" % (id, help))
			sys.exit(1)
		else:
			return defaultValue
	
name = getConfig("name", help="Descriptive name of the network.").strip()
shortName = getConfig("shortName", help="Shorter name of the network (for e.g. figures).").strip()
treeFile = os.path.join(dirConfigJson, getConfig("treeFile", help="Tree file containing the class hierarchy."))
assert os.path.exists(treeFile), treeFile
trainAnnotFile = os.path.join(dirConfigJson, getConfig("trainAnnotFile", help="File containing the annotated training sequences."))
testAnnotFile = os.path.join(dirConfigJson, getConfig("testAnnotFile", help="File containing the annotated test sequences."))
classBalancing = getConfig("classBalancing", 1, help="Whether to use class balancing or not (1 or 0).") != 0
outFilePrefix = os.path.join(dirConfigJson, getConfig("outFilePrefix", help="Filename prefix for output (e.g. checkpoint and log) files."))
classSizesCacheFile = outFilePrefix+"_class_sizes.cache.json"
ckptKeep = getConfig("ckptKeep", 1, help="Whether to keep all checkpoint files (0 or 1).") != 0
seqEncoder = eval("encode_x."+getConfig("seqEncoder", help="The encoder class (must be a class defined in <tt>encode_x.py</tt>)."))
minLength = getConfig("minLength", help="Minimum length of training/testing sequences to consider.")
maxClassInbalance = getConfig("maxClassInbalance", 5, help="Weight limit of rare classes when doing class balancing.")
maxLength = getConfig("maxLength", help="Maximum length of training/testing sequences to consider.")
batchSize = getConfig("batchSize", help="Minibatch size (number of sequences).")
lrate = getConfig("lrate", help="Learning rate.")
maxTestCount = getConfig("maxTestCount", 1000000, help="Maximum number of testing sequences.")
iters = getConfig("iters", help="Number of training iterations (minibatches).")
lrateDecay = getConfig("lrateDecay", 1, help="Learning rate decay parameter (1: no decay).")
lrateDecaySteps = getConfig("lrateDecaySteps", 1000000, help="Number of iterations (minibatches) after the learning rate is multiplied with the learning rate decay parameter.")
statStep = getConfig("statStep", 100, help="Number of iterations after which training statistics are printed.")
saveStep = getConfig("saveStep", 1000, help="Number of iterations after which a checkpoint (the model state) is saved.")
maxLevel = getConfig("maxLevel", help="Maximum level of classes to consider in the tree (>=0, 0 is the topmost level).")
minClassSize = getConfig("minClassSize", help="Minimum size of a class to consider (classes smaller than this size are dropped).")
csvSeparator = getConfig("csvSeparator", ",", help="Separator in CSV files. If set to a semicolon, then comma is used as the decimal separator.")
regLambda = getConfig("regLambda", 0, help="Weight decay parameter (multiplier for L2 regularization loss).")
removeStartingM = getConfig("removeStartingM", 1, help="Whether to remove the starting M amino acid from proteins (0 or 1).") != 0
network = getConfig("network", help="The definition of the network. The special values \"simple\" and \"complex\" correspond to built-in example network architectures.")

if network == "simple":
	# This is a simple network, suitable for classification into a few classes.
	print("Using %s example network" % network)
	network = [
		{"layer":"conv", "k":6, "d":32, "act":"prelu"},
		{"layer":"maxpool", "k":3}, 
		{"layer":"bn", "scope":"seq_batch_norm_0"},
		
		{"layer":"conv", "k":5, "d":64, "act":"prelu"},
		{"layer":"maxpool", "k":3}, 
		{"layer":"bn", "scope":"seq_batch_norm_1"},
		
		{"layer":"conv", "k":5, "d":128, "act":"prelu"},
		{"layer":"maxpool", "k":3}, 
		{"layer":"bn", "scope":"seq_batch_norm_2"},
		
		{"layer":"spp", "levels":1, "div":1}, # global max-pooling
		
		{"layer":"fc", "n":128, "act":"prelu"},
		{"layer":"dropout", "p":0.5, "scope":"fc0_dropout"},
		{"layer":"bn", "scale":1, "scope":"fc0_batch_norm"},
		
		{"layer":"fc", "n":"#out", "scope":"fcClassify"}
	]
	
elif network == "complex":
	# This is a more complex network, suitable for classification into a lot of classes.
	print("Using %s example network" % network)
	network = [
		{"layer":"conv", "k":6, "d":128, "act":"prelu"},
		{"layer":"bn", "scope":"seq_batch_norm_0"},
		
		{"layer":"conv", "k":6, "d":128, "act":"prelu"},
		{"layer":"maxpool", "k":2}, 
		{"layer":"bn", "scope":"seq_batch_norm_1"},
		
		{"layer":"conv", "k":5, "d":256, "act":"prelu"},
		{"layer":"maxpool", "k":2}, 
		{"layer":"bn", "scope":"seq_batch_norm_2"},
		
		{"layer":"conv", "k":5, "d":256, "act":"prelu"},
		{"layer":"maxpool", "k":2}, 
		{"layer":"bn", "scope":"seq_batch_norm_3"},
		
		{"layer":"conv", "k":5, "d":512, "act":"prelu"},
		{"layer":"maxpool", "k":2}, 
		{"layer":"bn", "scope":"seq_batch_norm_4"},
		
		{"layer":"conv", "k":5, "d":512, "act":"prelu"},
		{"layer":"maxpool", "k":2}, 
		{"layer":"bn", "scope":"seq_batch_norm_5"},
		
		{"layer":"spp", "levels":3, "div":4},
		
		{"layer":"fc", "n":1024, "act":"prelu"},
		{"layer":"dropout", "p":0.5, "scope":"fc0_dropout"},
		{"layer":"bn", "scale":1, "scope":"fc0_batch_norm"},
		
		{"layer":"fc", "n":"#out", "scope":"fcClassify"}
	]

configOk = True
for key in configJson:
	if key not in configDesc:
		print("Error: Unrecognized option '%s' in %s." % (key, fnConfigJson))
		configOk = False
if not configOk:
	sys.exit(1)

if "--html_help" in sys.argv: # print HTML help for config options
	print("<p>These are the possible keys for <em>config.json</em>. Configuration parameters without a default value are required, the remaining are optional:</p>")
	print("<ul>")
	for key in configDesc:
		print("<li>")
		print("<strong>%s</strong> (%s)" % (key, "default value: %s"%configDesc[key]["defaultValue"] if configDesc[key]["defaultValue"] is not None else "<strong>no default value</strong>"))
		print("<ul><li>")
		print(configDesc[key]["help"])
		print("</li></ul>")
		print("</li>")
	print("</ul>")
	sys.exit(1)
