#!/usr/bin/env python3

# SECLAF Sequence Classification Framework
# Copyright (c) 2016-2017 Balazs Szalkai

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os,sys,re,math,time
import numpy as np

from .seq_util import *
from .tree import *
from .annotation import *

# encode_x.py: Encoding and decoding sequences to/from numeric arrays, which can then be fed to Tensorflow to train or test the network.
# This file provides encoder classes for converting DNA and protein sequences to/from numpy ndarrays.

# Encodes each amino acid with a one-hot vector.
class SimpleAminoAcidEncoder:
	def __init__(self):
		pass
		
	# Gets the dimensions of the vector encoding a sequence element.
	def getDim(self):
		return len(AminoAcids)
		
	# Converts a protein sequence to a numpy ndarray, encoding each amino acid as a one-hot vector.
	# result: shape=[seq_length, 20]
	def encode(self, seq):
		protein_length = len(seq)
		np_prot = np.zeros([protein_length, self.getDim()], dtype=np.float32)
		for i in range(protein_length):
			aaIndex = getAminoAcidIndex(seq[i])
			if aaIndex >= 0:
				np_prot[i, aaIndex] = 1
		result = np_prot
		assert result.shape[1] == self.getDim()
		return result

	# Decodes a numpy ndarray as an amino acid (protein) sequence.
	def decode(self, np_protein):
		protein_length = np_protein.shape[0]
		np_protein = np.argmax(np_protein, 1)
		seq = ""
		for i in range(protein_length):
			seq += getAminoAcidFromIndex(np_protein[i])
		return seq
		
# Encodes amino acids in less dimensions by functional and structural properties.
class CompactAminoAcidEncoder:
	def __init__(self):
		pass
		
	# Gets the dimensions of the vector encoding a sequence element.
	def getDim(self):
		return 6
		
	# Converts a protein sequence to a numpy ndarray.
	# result: shape=[seq_length, 6]
	def encode(self, seq):
		protein_length = len(seq)
		np_prot = []
		for i in range(protein_length):
			a = seq[i]
			if getAminoAcidIndex(a) < 0: # unknown amino acid character
				descArray = [0.0] * 6
			else:
				descArray = [float(x) for x in [getAminoAcidCharge(a), getAminoAcidHydrophobicity(a), isAminoAcidPolar(a), isAminoAcidAromatic(a), hasAminoAcidHydroxyl(a), hasAminoAcidSulfur(a)]]
			np_prot.append(descArray)
			
		assert len(np_prot) > 0, "Error: Empty sequence"
		result = np.asfarray(np_prot)
		assert result.shape[1] == self.getDim()
		return result

	# Decodes a numpy ndarray as an amino acid (protein) sequence.
	def decode(self, npy):
		raise NotImplementedError("Currently cannot decode compact amino acid representation")
	
# Combines the simple and compact encoders.
class BigAminoAcidEncoder:
	def __init__(self):
		self.simple = SimpleAminoAcidEncoder()
		self.compact = CompactAminoAcidEncoder()
		
	# Gets the dimensions of the vector encoding a sequence element.
	def getDim(self):
		return self.simple.getDim() + self.compact.getDim()
		
	# Converts a protein sequence to a numpy ndarray.
	# result: shape=[seq_length, 26]
	def encode(self, seq):
		a1 = self.simple.encode(seq)
		a2 = self.compact.encode(seq)
		result = np.concatenate((a1, a2), 1)
		assert result.shape[1] == self.getDim()
		return result

	# Decodes a numpy ndarray as an amino acid (protein) sequence.
	def decode(self, npy):
		return self.simple.decode(npy[:,:self.simple.getDim()])
	
# Encodes nucleotides with 4 "bits" each.
class SimpleDnaEncoder:
	def __init__(self):
		pass
		
	# Gets the dimensions of the vector encoding a sequence element.
	def getDim(self):
		return 4
		
	# Converts a DNA sequence to a numpy ndarray.
	# result: shape=[seq_length, 4]
	def encode(self, seq):
		dna_length = len(seq)
		np_dna = np.zeros([dna_length, self.getDim()], dtype=np.float32)
		for i in range(dna_length):
			dnaIndex = getDnaIndex(seq[i])
			if dnaIndex >= 0:
				np_dna[i, dnaIndex] = 1
		assert np_dna.shape[1] == self.getDim()
		return np_dna
			
	# Decodes a numpy ndarray as a DNA sequence.
	def decode(self, np_dna):
		dna_length = np_dna.shape[0]
		np_dna = np.argmax(np_dna, 1)
		seq = ""
		for i in range(dna_length):
			seq += getDnaFromIndex(np_dna[i])
		return seq

# Encodes nucleotides with 3 "bits" each: purine/pyrimidine, strong/weak, amino/keto.
class CompactDnaEncoder:
	def __init__(self):
		pass
		
	# Gets the dimensions of the vector encoding a sequence element.
	def getDim(self):
		return 3

	# Converts a DNA sequence to a numpy ndarray.
	# result: shape=[seq_length, 3]
	def encode(self, seq):
		dna_length = len(seq)
		np_dna = []
		for i in range(dna_length):
			n = seq[i]
			np_dna.append([float(n in "AG"), float(n in "CG"), float(n in "AC")])
		np_dna = np.asfarray(np_dna)
		assert np_dna.shape[1] == self.getDim()
		return np_dna
			
	# Decodes a numpy ndarray as a DNA sequence.
	def decode(self, np_dna):
		raise NotImplementedError("Currently cannot decode compact DNA representation")
